package demo.prague.markup.workspace;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextArea;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultHighlighter;
import javax.swing.text.Highlighter;
import javax.swing.text.Highlighter.HighlightPainter;

import ro.sync.ecss.extensions.api.AuthorAccess;
import ro.sync.ecss.extensions.api.AuthorDocumentController;
import ro.sync.ecss.extensions.api.AuthorOperationException;
import ro.sync.ecss.extensions.api.highlights.AuthorHighlighter;
import ro.sync.ecss.extensions.api.highlights.Highlight;
import ro.sync.ecss.extensions.api.highlights.HighlightActionsProvider;
import ro.sync.ecss.extensions.api.highlights.RectangleHighlightPainter;
import ro.sync.ecss.extensions.api.node.AuthorDocumentFragment;
import ro.sync.ecss.extensions.api.node.AuthorNode;
import ro.sync.exml.plugin.workspace.WorkspaceAccessPluginExtension;
import ro.sync.exml.workspace.api.PluginWorkspace;
import ro.sync.exml.workspace.api.editor.WSEditor;
import ro.sync.exml.workspace.api.editor.page.WSEditorPage;
import ro.sync.exml.workspace.api.editor.page.author.WSAuthorEditorPage;
import ro.sync.exml.workspace.api.editor.page.text.TextPopupMenuCustomizer;
import ro.sync.exml.workspace.api.editor.page.text.WSTextEditorPage;
import ro.sync.exml.workspace.api.editor.page.text.xml.WSXMLTextEditorPage;
import ro.sync.exml.workspace.api.editor.page.text.xml.WSXMLTextNodeRange;
import ro.sync.exml.workspace.api.editor.page.text.xml.XPathException;
import ro.sync.exml.workspace.api.listeners.WSEditorChangeListener;
import ro.sync.exml.workspace.api.listeners.WSEditorListener;
import ro.sync.exml.workspace.api.standalone.StandalonePluginWorkspace;
import ro.sync.exml.workspace.api.standalone.ToolbarComponentsCustomizer;
import ro.sync.exml.workspace.api.standalone.ToolbarInfo;
import ro.sync.exml.workspace.api.standalone.ViewComponentCustomizer;
import ro.sync.exml.workspace.api.standalone.ViewInfo;
import ro.sync.exml.workspace.api.standalone.ui.ToolbarButton;

/**
 * Plugin extension - workspace access extension.
 */
public class CustomWorkspaceAccessPluginExtension implements WorkspaceAccessPluginExtension {

  private static HashMap<String, String> terms = new HashMap<String, String>();
  private StandalonePluginWorkspace pluginWorkspaceAccess;
  private TextPopupMenuCustomizer textCustomizer;
  static {
    terms.put("Amsterdam", "http://en.wikipedia.org/wiki/Amsterdam");
    terms.put("Prague", "http://en.wikipedia.org/wiki/Prague");
  }
  
  @Override
  public void applicationStarted(final StandalonePluginWorkspace pluginWorkspaceAccess) {
    this.pluginWorkspaceAccess = pluginWorkspaceAccess;
    pluginWorkspaceAccess.addViewComponentCustomizer(new ViewComponentCustomizer() {
      @Override
      public void customizeView(ViewInfo viewInfo) {
        if ("SampleWorkspaceAccessID".equals(viewInfo.getViewID())) {
          JPanel panel = new JPanel(new BorderLayout());
          
          final JList entriesList = new JList(terms.keySet().toArray());
          
          entriesList.addMouseListener(new MouseAdapter() {
            @Override
            public void mouseClicked(MouseEvent e) {
              if (e.getClickCount() == 2 && !e.isPopupTrigger()) {
                String selectedValue = (String) entriesList.getSelectedValue();
                if (selectedValue != null) {
                  insertEntry(selectedValue);
                }
              }
            }
          });
          
          panel.add(entriesList, BorderLayout.CENTER);
          
          JButton HighlightButton = new JButton("Highlight refs");
          HighlightButton.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
              highlight();
            }
          });
          
          panel.add(HighlightButton, BorderLayout.SOUTH);
          
          viewInfo.setComponent(panel);
        }
      }
    });
    
    pluginWorkspaceAccess.addToolbarComponentsCustomizer(new ToolbarComponentsCustomizer() {
      @Override
      public void customizeToolbar(ToolbarInfo toolbarInfo) {
        if ("SampleWorkspaceAccessToolbarID".equals(toolbarInfo.getToolbarID())) {
          Action action = new AbstractAction() {
            @Override
            public void actionPerformed(ActionEvent arg0) {
              pluginWorkspaceAccess.showView("SampleWorkspaceAccessID", false);
            }
          };
          action.putValue(Action.NAME, "Show helper");
          
          ToolbarButton button = new ToolbarButton(action, true);
          
          toolbarInfo.setComponents(new JComponent[] {button});
        }
      }
    });
    
    pluginWorkspaceAccess.addEditorChangeListener(new WSEditorChangeListener() {
      @Override
      public void editorOpened(URL editorLocation) {
        final WSEditor editorAccess = pluginWorkspaceAccess.getEditorAccess(editorLocation, PluginWorkspace.MAIN_EDITING_AREA);
        install(editorAccess);
        editorAccess.addEditorListener(new WSEditorListener() {
          @Override
          public void editorPageChanged() {
            install(editorAccess);
          }
        });
      }
    }, PluginWorkspace.MAIN_EDITING_AREA);
  }
  
  protected void install(WSEditor editorAccess) {
    WSEditorPage currentPage = editorAccess.getCurrentPage();
    if (currentPage instanceof WSXMLTextEditorPage) {
      if (textCustomizer == null) {
        textCustomizer = new TextPopupMenuCustomizer() {
          @Override
          public void customizePopUpMenu(Object popUp, WSTextEditorPage textPage) {
            JPopupMenu jPopupMenu = (JPopupMenu) popUp;
            Set<String> keySet = terms.keySet();
            for (final String term : keySet) {
              Action action = new AbstractAction() {
                @Override
                public void actionPerformed(ActionEvent arg0) {
                  insertEntry(term);
                }
              };
              action.putValue(Action.NAME, term);
              jPopupMenu.insert(action, 0);
            }
          }
        };
      }
      WSXMLTextEditorPage textPage = (WSXMLTextEditorPage) currentPage;
      textPage.removePopUpMenuCustomizer(textCustomizer);
      textPage.addPopUpMenuCustomizer(textCustomizer);
    }
  }

  private void highlight() {
    WSEditor editorAccess = pluginWorkspaceAccess.getCurrentEditorAccess(PluginWorkspace.MAIN_EDITING_AREA);
    if (editorAccess != null) {
      WSEditorPage currentPage = editorAccess.getCurrentPage();
      
      if (currentPage instanceof WSXMLTextEditorPage) {
        WSXMLTextEditorPage textPage = (WSXMLTextEditorPage) currentPage;
        
        JTextArea textComponent = (JTextArea) textPage.getTextComponent();
        Highlighter highlighter = textComponent.getHighlighter();
        highlighter.removeAllHighlights();
        try {
          WSXMLTextNodeRange[] ranges = textPage.findElementsByXPath("//xref");
          HighlightPainter painter = 
              new DefaultHighlighter.DefaultHighlightPainter(Color.PINK);
          for (int i = 0; i < ranges.length; i++) {
            WSXMLTextNodeRange cRange = ranges[i];
            int startOffset = textPage.getOffsetOfLineStart(cRange.getStartLine()) + 
                cRange.getStartColumn();
            int endOffset = textPage.getOffsetOfLineStart(cRange.getEndLine()) + 
                cRange.getEndColumn();
            
            highlighter.addHighlight(startOffset, endOffset, painter);
          }
        } catch (XPathException e) {
          e.printStackTrace();
        } catch (BadLocationException e) {
          e.printStackTrace();
        }
        
      } else if (currentPage instanceof WSAuthorEditorPage) {
        final WSAuthorEditorPage authorPage = (WSAuthorEditorPage) currentPage;
        
        final AuthorDocumentController ctrl = authorPage.getDocumentController();
        try {
          AuthorNode[] nodes = ctrl.findNodesByXPath("//xref", true, true, true);
          AuthorHighlighter highlighter = authorPage.getHighlighter();
          highlighter.removeAllHighlights();
          
          ro.sync.ecss.extensions.api.highlights.HighlightPainter painter = 
              new RectangleHighlightPainter(ro.sync.exml.view.graphics.Color.COLOR_BLUE);
          for (int i = 0; i < nodes.length; i++) {
            final AuthorNode cNode = nodes[i];
            Highlight addHighlight = highlighter.addHighlight(
                cNode.getStartOffset(), 
                cNode.getEndOffset(), 
                painter, 
                null);
            
            HighlightActionsProvider additionalDataProvider = new HighlightActionsProvider() {
              @Override
              public Object getExpandedWidgetIcon() {
                return new ImageIcon(getClass().getResource("/images/ShowHelp16.png"));
              }
              
              @Override
              public Object getCollapsedWidgetIcon() {
                return getExpandedWidgetIcon();
              }
              
              @Override
              public List getActions(AuthorAccess authorAccess) {
                List actions = new ArrayList();
                
                Action removeXref = new AbstractAction() {
                  @Override
                  public void actionPerformed(ActionEvent e) {
                    try {
                      ctrl.beginCompoundEdit();
                      int startOffset = cNode.getStartOffset();
                      AuthorDocumentFragment frag = ctrl.createDocumentFragment(startOffset + 1, cNode.getEndOffset() - 1);
                      ctrl.deleteNode(cNode);
                      ctrl.insertFragment(startOffset, frag);
                      authorPage.select(startOffset, startOffset + frag.getLength());
                    } catch (BadLocationException e1) {
                      e1.printStackTrace();
                    } finally {
                      ctrl.endCompoundEdit();
                    }
                  }
                };
                removeXref.putValue(Action.NAME, "Unwrap");
                actions.add(removeXref);
                
                return actions;
              }
            };
            
            addHighlight.setAdditionalData(Highlight.HOVER_ACTIONS_PROVIDER_ADDITIONAL_DATA, additionalDataProvider);
          }
        } catch (AuthorOperationException e) {
          e.printStackTrace();
        } catch (BadLocationException e) {
          e.printStackTrace();
        }
      }
    }
  }
  
  private void insertEntry(String term) {
    WSEditor editorAccess = pluginWorkspaceAccess.getCurrentEditorAccess(PluginWorkspace.MAIN_EDITING_AREA);
    if (editorAccess != null) {
      WSEditorPage currentPage = editorAccess.getCurrentPage();
      String fragment = 
          "<xref href=\"" + terms.get(term) + "\" format=\"html\">" 
              + term 
              + "</xref>";
      if (currentPage instanceof WSXMLTextEditorPage) {
        WSXMLTextEditorPage textPage = (WSXMLTextEditorPage) currentPage;
        
        JTextArea textComponent = (JTextArea) textPage.getTextComponent();
        textComponent.insert(fragment, textComponent.getCaretPosition());
        
        textComponent.requestFocus();
      } else if (currentPage instanceof WSAuthorEditorPage) {
        WSAuthorEditorPage authorPage = (WSAuthorEditorPage) currentPage;
        AuthorDocumentController ctrl = authorPage.getDocumentController();
        
        try {
          ctrl.insertXMLFragment(fragment, authorPage.getCaretOffset());
          
          AuthorNode nodeAtOffset = ctrl.getNodeAtOffset(authorPage.getCaretOffset());
          authorPage.setCaretPosition(nodeAtOffset.getEndOffset() + 1);
          
          JComponent authorComponent = (JComponent) authorPage.getAuthorComponent();
          authorComponent.requestFocus();
        } catch (AuthorOperationException e) {
          e.printStackTrace();
        } catch (BadLocationException e) {
          e.printStackTrace();
        }
      }
    }
  }

  @Override
  public boolean applicationClosing() {
    return true;
  }
  
}